@@ -1,6 +1,6 @@
 /********************************************
 bi_funct.c
-copyright 2008-2012,2013, Thomas E. Dickey
+copyright 2008-2014,2016, Thomas E. Dickey
 copyright 1991-1995,1996, Michael D. Brennan
 
 This is a source file for mawk, an implementation of
@@ -11,7 +11,7 @@ the GNU General Public License, version 2, 1991.
 ********************************************/
 
 /*
- * $MawkId: bi_funct.c,v 1.67 2013/02/19 10:57:28 tom Exp $
+ * $MawkId: bi_funct.c,v 1.105 2016/06/16 00:36:50 tom Exp $
  * @Log: bi_funct.c,v @
  * Revision 1.9  1996/01/14  17:16:11  mike
  * flush_all_output() before system()
@@ -78,7 +78,17 @@ the GNU General Public License, version 2, 1991.
 #include <math.h>
 #include <time.h>
 
-/* #define EXP_UNROLLED_GSUB */
+#if defined(mawk_srand) || defined(mawk_rand)
+#define USE_SYSTEM_SRAND
+#endif
+
+#if defined(HAVE_BSD_STDLIB_H) && defined(USE_SYSTEM_SRAND)
+#include <bsd/stdlib.h>		/* prototype arc4random */
+#endif
+
+#if defined(WINVER) && (WINVER >= 0x501)
+#include <windows.h>
+#endif
 
 #if OPT_TRACE > 0
 #define return_CELL(func, cell) TRACE(("..." func " ->")); \
@@ -146,10 +156,11 @@ bi_funct_init(void)
 #ifndef NO_INIT_SRAND
     /* seed rand() off the clock */
     {
-	CELL c;
+	CELL c[2];
 
-	c.type = C_NOINIT;
-	bi_srand(&c);
+	memset(c, 0, sizeof(c));
+	c[1].type = C_NOINIT;
+	bi_srand(c + 1);
     }
 #endif
 
@@ -160,7 +171,7 @@ bi_funct_init(void)
  **************************************************/
 
 CELL *
-bi_length(CELL * sp)
+bi_length(CELL *sp)
 {
     size_t len;
 
@@ -227,7 +238,7 @@ str_str(char *target, size_t target_len, char *key, size_t key_len)
 }
 
 CELL *
-bi_index(CELL * sp)
+bi_index(CELL *sp)
 {
     size_t idx;
     size_t len;
@@ -262,7 +273,7 @@ bi_index(CELL * sp)
     from  max(1,i) to min(l,n-i-1) inclusive */
 
 CELL *
-bi_substr(CELL * sp)
+bi_substr(CELL *sp)
 {
     int n_args, len;
     register int i, n;
@@ -316,7 +327,7 @@ bi_substr(CELL * sp)
     }
 
     /*
-     * Keep 'n' from extending past the end of the string. 
+     * Keep 'n' from extending past the end of the string.
      */
     if (n > len - i) {
 	n = len - i;
@@ -341,7 +352,7 @@ bi_substr(CELL * sp)
 */
 
 CELL *
-bi_match(CELL * sp)
+bi_match(CELL *sp)
 {
     char *p;
     size_t length;
@@ -358,7 +369,11 @@ bi_match(CELL * sp)
     RSTART->type = C_DOUBLE;
     RLENGTH->type = C_DOUBLE;
 
-    p = REmatch(string(sp)->str, string(sp)->len, cast_to_re((sp + 1)->ptr), &length);
+    p = REmatch(string(sp)->str,
+		string(sp)->len,
+		cast_to_re((sp + 1)->ptr),
+		&length,
+		0);
 
     if (p) {
 	sp->dval = (double) (p - string(sp)->str + 1);
@@ -377,7 +392,7 @@ bi_match(CELL * sp)
 }
 
 CELL *
-bi_toupper(CELL * sp)
+bi_toupper(CELL *sp)
 {
     STRING *old;
     register char *p, *q;
@@ -401,7 +416,7 @@ bi_toupper(CELL * sp)
 }
 
 CELL *
-bi_tolower(CELL * sp)
+bi_tolower(CELL *sp)
 {
     STRING *old;
     register char *p, *q;
@@ -428,7 +443,7 @@ bi_tolower(CELL * sp)
  * Like gawk...
  */
 CELL *
-bi_systime(CELL * sp)
+bi_systime(CELL *sp)
 {
     time_t result;
     time(&result);
@@ -443,12 +458,12 @@ bi_systime(CELL * sp)
 
 #ifdef HAVE_MKTIME
 /*  mktime(datespec)
-    Turns datespec into a time stamp of the same form as returned by systime(). 
+    Turns datespec into a time stamp of the same form as returned by systime().
     The datespec is a string of the form
         YYYY MM DD HH MM SS [DST].
 */
 CELL *
-bi_mktime(CELL * sp)
+bi_mktime(CELL *sp)
 {
     time_t result;
     struct tm my_tm;
@@ -492,7 +507,7 @@ bi_mktime(CELL * sp)
 }
 #endif
 
-/*  strftime(format, timestamp, utc) 
+/*  strftime(format, timestamp, utc)
     should be equal to gawk strftime. all parameters are optional:
         format: ansi c strftime format descriptor. default is "%c"
         timestamp: seconds since unix epoch. default is now
@@ -500,7 +515,7 @@ bi_mktime(CELL * sp)
 */
 #ifdef HAVE_STRFTIME
 CELL *
-bi_strftime(CELL * sp)
+bi_strftime(CELL *sp)
 {
     const char *format = "%c";
     time_t rawtime;
@@ -591,7 +606,7 @@ fplib_err(
 #endif
 
 CELL *
-bi_sin(CELL * sp)
+bi_sin(CELL *sp)
 {
     TRACE_FUNC("bi_sin", sp);
 
@@ -616,7 +631,7 @@ bi_sin(CELL * sp)
 }
 
 CELL *
-bi_cos(CELL * sp)
+bi_cos(CELL *sp)
 {
     TRACE_FUNC("bi_cos", sp);
 
@@ -641,7 +656,7 @@ bi_cos(CELL * sp)
 }
 
 CELL *
-bi_atan2(CELL * sp)
+bi_atan2(CELL *sp)
 {
     TRACE_FUNC("bi_atan2", sp);
 
@@ -665,7 +680,7 @@ bi_atan2(CELL * sp)
 }
 
 CELL *
-bi_log(CELL * sp)
+bi_log(CELL *sp)
 {
     TRACE_FUNC("bi_log", sp);
 
@@ -690,7 +705,7 @@ bi_log(CELL * sp)
 }
 
 CELL *
-bi_exp(CELL * sp)
+bi_exp(CELL *sp)
 {
     TRACE_FUNC("bi_exp", sp);
 
@@ -716,7 +731,7 @@ bi_exp(CELL * sp)
 }
 
 CELL *
-bi_int(CELL * sp)
+bi_int(CELL *sp)
 {
     TRACE_FUNC("bi_int", sp);
 
@@ -727,7 +742,7 @@ bi_int(CELL * sp)
 }
 
 CELL *
-bi_sqrt(CELL * sp)
+bi_sqrt(CELL *sp)
 {
     TRACE_FUNC("bi_sqrt", sp);
 
@@ -751,9 +766,7 @@ bi_sqrt(CELL * sp)
     return_CELL("bi_sqrt", sp);
 }
 
-#if defined(mawk_srand) || defined(mawk_rand)
-#define USE_SYSTEM_SRAND
-#else
+#if !(defined(mawk_srand) || defined(mawk_rand))
 /* For portability, we'll use our own random number generator , taken
    from:  Park, SK and Miller KW, "Random Number Generators:
    Good Ones are Hard to Find", CACM, 31, 1192-1201, 1988.
@@ -781,8 +794,32 @@ static CELL cseed;		/* argument of last call to srand() */
 #endif /* M == MAX__LONG */
 #endif /* defined(mawk_srand) || defined(mawk_rand) */
 
+static double
+initial_seed(void)
+{
+    double result;
+#if defined(HAVE_GETTIMEOFDAY)
+    struct timeval data;
+    gettimeofday(&data, (struct timezone *) 0);
+    result = (data.tv_sec * 1000000) + data.tv_usec;
+#elif defined(WINVER) && (WINVER >= 0x501)
+    union {
+	FILETIME ft;
+	long long since1601;	/* time since 1 Jan 1601 in 100ns units */
+    } data;
+
+    GetSystemTimeAsFileTime(&data.ft);
+    result = (double) (data.since1601 / 10LL);
+#else
+    time_t now;
+    (void) time(&now);
+    result = (double) now;
+#endif
+    return result;
+}
+
 CELL *
-bi_srand(CELL * sp)
+bi_srand(CELL *sp)
 {
 #ifdef USE_SYSTEM_SRAND
     static long seed = 1;
@@ -798,11 +835,10 @@ bi_srand(CELL * sp)
 
     if (sp->type == C_NOINIT)	/* seed off clock */
     {
-	time_t secs = time((time_t *) 0);
 	cellcpy(sp, &cseed);
 	cell_destroy(&cseed);
 	cseed.type = C_DOUBLE;
-	cseed.dval = (double) secs;
+	cseed.dval = initial_seed();
     } else {			/* user seed */
 	sp--;
 	/* swap cseed and *sp ; don't need to adjust ref_cnts */
@@ -839,7 +875,7 @@ bi_srand(CELL * sp)
 }
 
 CELL *
-bi_rand(CELL * sp)
+bi_rand(CELL *sp)
 {
     TRACE_FUNC("bi_rand", sp);
 
@@ -848,7 +884,7 @@ bi_rand(CELL * sp)
 	long value = (long) mawk_rand();
 	sp++;
 	sp->type = C_DOUBLE;
-	sp->dval = ((double) value) / RAND_MAX;
+	sp->dval = ((double) value) / MAWK_RAND_MAX;
     }
 #else
     crank(seed);
@@ -874,7 +910,7 @@ bi_rand(CELL * sp)
  *************************************************/
 
 CELL *
-bi_close(CELL * sp)
+bi_close(CELL *sp)
 {
     int x;
 
@@ -891,7 +927,7 @@ bi_close(CELL * sp)
 }
 
 CELL *
-bi_fflush(CELL * sp)
+bi_fflush(CELL *sp)
 {
     int ret = 0;
 
@@ -914,7 +950,7 @@ bi_fflush(CELL * sp)
 }
 
 CELL *
-bi_system(CELL * sp GCC_UNUSED)
+bi_system(CELL *sp GCC_UNUSED)
 {
 #ifdef HAVE_REAL_PIPES
     int pid;
@@ -972,7 +1008,7 @@ bi_system(CELL * sp GCC_UNUSED)
 */
 
 CELL *
-bi_getline(CELL * sp)
+bi_getline(CELL *sp)
 {
     CELL tc;
     CELL *cp = 0;
@@ -1082,7 +1118,7 @@ bi_getline(CELL * sp)
 */
 
 CELL *
-bi_sub(CELL * sp)
+bi_sub(CELL *sp)
 {
     CELL *cp;			/* pointer to the replacement target */
     CELL tc;			/* build the new string here */
@@ -1105,7 +1141,13 @@ bi_sub(CELL * sp)
 	cast1_to_s(&sc);
     front = string(&sc)->str;
 
-    if ((middle = REmatch(front, string(&sc)->len, cast_to_re(sp->ptr), &middle_len))) {
+    middle = REmatch(front,
+		     string(&sc)->len,
+		     cast_to_re(sp->ptr),
+		     &middle_len,
+		     0);
+
+    if (middle != 0) {
 	front_len = (size_t) (middle - front);
 	back = middle + middle_len;
 	back_len = string(&sc)->len - front_len - middle_len;
@@ -1149,320 +1191,164 @@ bi_sub(CELL * sp)
     return_CELL("bi_sub", sp);
 }
 
-typedef enum {
-    btFinish = 0,
-    btNormal,
-    btEmpty
-} GSUB_BT;
-
-typedef struct {
-    STRING *result;
-    CELL replace;
-    char *front;
-    char *middle;
-    char *target;
-    size_t front_len;
-    size_t middle_len;
-    size_t target_len;
-    int empty_ok;
-    GSUB_BT branch_to;
-} GSUB_STK;
-
-#define ThisGSUB       gsub_stk[level]
-#define ThisBranch     ThisGSUB.branch_to
-#define ThisEmptyOk    ThisGSUB.empty_ok
-#define ThisFront      ThisGSUB.front
-#define ThisFrontLen   ThisGSUB.front_len
-#define ThisMiddle     ThisGSUB.middle
-#define ThisMiddleLen  ThisGSUB.middle_len
-#define ThisReplace    ThisGSUB.replace
-#define ThisResult     ThisGSUB.result
-#define ThisTarget     ThisGSUB.target
-#define ThisTargetLen  ThisGSUB.target_len
-
-#define NextGSUB       gsub_stk[level + 1]
-#define NextBranch     NextGSUB.branch_to
-#define NextEmptyOk    NextGSUB.empty_ok
-#define NextFront      NextGSUB.front
-#define NextFrontLen   NextGSUB.front_len
-#define NextMiddle     NextGSUB.middle
-#define NextMiddleLen  NextGSUB.middle_len
-#define NextReplace    NextGSUB.replace
-#define NextResult     NextGSUB.result
-#define NextTarget     NextGSUB.target
-#define NextTargetLen  NextGSUB.target_len
-
-#ifdef EXP_UNROLLED_GSUB
-/* #define DEBUG_GSUB 1 */
-
-static size_t gsub_max;
-static GSUB_STK *gsub_stk;
 static unsigned repl_cnt;	/* number of global replacements */
 
-#if OPT_TRACE > 0
-static const char *
-indent(int level)
-{
-    static const char value[] = "-----------------";
-    const char *result;
-    int limit = (int) sizeof(value) - 1;
-
-    if (level < limit)
-	result = value + limit - level;
-    else
-	result = "";
-    return result;
-}
-#endif
-
-/* recursive global subsitution
-   dealing with empty matches makes this mildly painful
-
-   repl is always of type REPL or REPLV, destroyed by caller
-   empty_ok is set if, match of empty string at front is OK
-*/
-
-#ifdef DEBUG_GSUB
 static STRING *
-old_gsub(PTR re, int level)
+gsub3(PTR re, CELL *repl, CELL *target)
 {
-    char xbuff[2];
-    char *in_sval;
-    char *front = 0, *middle;
-    STRING *back;
-    size_t front_len, middle_len;
+    int j;
+    CELL xrepl;
+    STRING *input = string(target);
+    STRING *output;
+    STRING *buffer;
+    STRING *sval;
+    size_t have;
+    size_t used = 0;
+    size_t guess = input->len;
+    size_t limit = guess;
 
-    assert(level >= 0);
-    assert(level + 1 < (int) gsub_max);
+    int skip0 = -1;
+    size_t howmuch;
+    char *where;
 
-    middle = REmatch(ThisTarget, ThisTargetLen, cast_to_re(re), &middle_len);
-    if (middle != 0) {
-
-	if (!ThisEmptyOk && (middle_len == 0) && (middle == ThisTarget)) {
-	    /* match at front that's not allowed */
-
-	    front_len = 0;
-
-	    if (ThisTargetLen == 0) {	/* target is empty string */
-		null_str.ref_cnt++;
-		back = &null_str;
-	    } else if (isAnchored(re)) {
-		back = new_STRING1(ThisTarget, ThisTargetLen);
-	    } else {
-		/* make new repl with target[0] */
-		cellcpy(&NextReplace, &ThisReplace);
-		repl_destroy(&ThisReplace);
-		xbuff[0] = *ThisTarget;
-		xbuff[1] = 0;
-		ThisReplace.type = C_REPL;
-		ThisReplace.ptr = (PTR) new_STRING1(xbuff, (size_t) 1);
-
-		NextTarget = ThisTarget + 1;
-		NextTargetLen = ThisTargetLen - 1;
-		NextEmptyOk = 1;
-		NextBranch = btEmpty;
-
-		back = old_gsub(re, level + 1);
-	    }
-	} else {		/* a match that counts */
-	    repl_cnt++;
-
-	    front = ThisTarget;
-	    front_len = (size_t) (middle - ThisTarget);
-
-	    if (front_len == ThisTargetLen) {	/* matched back of target */
-		back = &null_str;
-		null_str.ref_cnt++;
-	    } else {
-		NextTarget = middle + middle_len;
-		NextTargetLen = ThisTargetLen - (front_len + middle_len);
-		NextEmptyOk = 0;
-		NextBranch = btNormal;
-		cellcpy(&NextReplace, &ThisReplace);
-
-		back = old_gsub(re, level + 1);
-	    }
-
-	    /* patch the &'s if needed */
-	    if (ThisReplace.type == C_REPLV) {
-		STRING *sval = new_STRING1(middle, middle_len);
-
-		replv_to_repl(&ThisReplace, sval);
-		free_STRING(sval);
-	    }
-	}
-
-	/* put the three pieces together */
-	ThisResult = new_STRING0(front_len + string(&ThisReplace)->len + back->len);
-	TRACE(("old %s front '%.*s', middle '%.*s', back '%.*s'\n",
-	       indent(level),
-	       front_len, front,
-	       string(&ThisReplace)->len, string(&ThisReplace)->str,
-	       back->len, back->str));
-	in_sval = ThisResult->str;
-
-	if (front_len) {
-	    memcpy(in_sval, front, front_len);
-	    in_sval += front_len;
-	}
-	if (string(&ThisReplace)->len) {
-	    memcpy(in_sval, string(&ThisReplace)->str, string(&ThisReplace)->len);
-	    in_sval += string(&ThisReplace)->len;
-	}
-	if (back->len)
-	    memcpy(in_sval, back->str, back->len);
-
-	/* cleanup, repl is freed by the caller */
-	free_STRING(back);
+    TRACE(("called gsub3\n"));
 
+    /*
+     * If the replacement is constant, do it only once.
+     */
+    if (repl->type != C_REPLV) {
+	cellcpy(&xrepl, repl);
     } else {
-	/* no match */
-	ThisResult = new_STRING1(ThisTarget, ThisTargetLen);
+	memset(&xrepl, 0, sizeof(xrepl));
     }
 
-    return ThisResult;
-}
-#endif /* DEBUG_GSUB */
+    repl_cnt = 0;
+    output = new_STRING0(limit);
 
-static STRING *
-new_gsub(PTR re, int level)
-{
-    char xbuff[2];
-    char *in_sval;
-    STRING *back;
+    for (j = 0; j <= (int) input->len; ++j) {
+	where = REmatch(input->str + j,
+			input->len - (size_t) j,
+			cast_to_re(re),
+			&howmuch,
+			(j != 0));
+	/*
+	 * REmatch returns a non-null pointer if it found a match.  But
+	 * that can be an empty string, e.g., for "*" or "?".  The length
+	 * is in 'howmuch'.
+	 */
+	if (where != 0) {
+	    have = (size_t) (where - (input->str + j));
+	    if (have) {
+		skip0 = -1;
+		TRACE(("..before match:%d:", (int) have));
+		TRACE_STRING2(input->str + j, have);
+		TRACE(("\n"));
+		memcpy(output->str + used, input->str + j, have);
+		used += have;
+	    }
 
-  loop:
-    assert(level >= 0);
-    assert(level + 1 < (int) gsub_max);
+	    TRACE(("REmatch %d vs %d len=%d:", (int) j, skip0, (int) howmuch));
+	    TRACE_STRING2(where, howmuch);
+	    TRACE(("\n"));
 
-    ThisFront = 0;
+	    if (repl->type == C_REPLV) {
+		if (xrepl.ptr == 0 ||
+		    string(&xrepl)->len != howmuch ||
+		    (howmuch != 0 &&
+		     memcmp(string(&xrepl)->str, where, howmuch))) {
+		    if (xrepl.ptr != 0)
+			repl_destroy(&xrepl);
+		    sval = new_STRING1(where, howmuch);
+		    cellcpy(&xrepl, repl);
+		    replv_to_repl(&xrepl, sval);
+		    free_STRING(sval);
+		}
+	    }
 
-    ThisMiddle = REmatch(ThisTarget, ThisTargetLen, cast_to_re(re), &ThisMiddleLen);
-    if (ThisMiddle != 0) {
+	    have = string(&xrepl)->len;
+	    TRACE(("..replace:"));
+	    TRACE_STRING2(string(&xrepl)->str, have);
+	    TRACE(("\n"));
 
-	if (!ThisEmptyOk && (ThisMiddleLen == 0) && (ThisMiddle == ThisTarget)) {
-	    /* match at front that's not allowed */
+	    if (howmuch || (j != skip0)) {
+		++repl_cnt;
 
-	    ThisFrontLen = 0;
+		/*
+		 * If this new chunk is longer than its replacement, add that
+		 * to the estimate of the length.  Then, if the estimate goes
+		 * past the allocated length, reallocate and copy the existing
+		 * data.
+		 */
+		if (have > howmuch) {	/* growing */
+		    guess += (have - howmuch);
+		    if (guess >= limit) {
+			buffer = output;
+			limit = (++guess) * 2;	/* FIXME - too coarse? */
+			output = new_STRING0(limit);
+			memcpy(output->str, buffer->str, used);
+			free_STRING(buffer);
+		    }
+		} else if (howmuch > have) {	/* shrinking */
+		    guess -= (howmuch - have);
+		}
 
-	    if (ThisTargetLen == 0) {	/* target is empty string */
-		null_str.ref_cnt++;
-		back = &null_str;
-	    } else if (isAnchored(re)) {
-		back = new_STRING1(ThisTarget, ThisTargetLen);
+		/*
+		 * Finally, copy the new chunk.
+		 */
+		memcpy(output->str + used, string(&xrepl)->str, have);
+		used += have;
+	    }
+
+	    if (howmuch) {
+		j = (int) ((size_t) (where - input->str) + howmuch) - 1;
 	    } else {
-		/* make new repl with target[0] */
-		cellcpy(&NextReplace, &ThisReplace);
-		repl_destroy(&ThisReplace);
-		xbuff[0] = *ThisTarget;
-		xbuff[1] = 0;
-		ThisReplace.type = C_REPL;
-		ThisReplace.ptr = (PTR) new_STRING1(xbuff, (size_t) 1);
-
-		NextTarget = ThisTarget + 1;
-		NextTargetLen = ThisTargetLen - 1;
-		NextEmptyOk = 1;
-		NextBranch = btEmpty;
-
-		++level;
-		goto loop;
-
-	      empty_match:
-		back = NextResult;
-	    }
-	} else {		/* a match that counts */
-	    repl_cnt++;
-
-	    ThisFront = ThisTarget;
-	    ThisFrontLen = (size_t) (ThisMiddle - ThisTarget);
-
-	    if (ThisFrontLen == ThisTargetLen) {	/* matched back of target */
-		back = &null_str;
-		null_str.ref_cnt++;
-	    } else {
-		NextTarget = ThisMiddle + ThisMiddleLen;
-		NextTargetLen = ThisTargetLen - (ThisFrontLen + ThisMiddleLen);
-		NextEmptyOk = 0;
-		NextBranch = btNormal;
-		cellcpy(&NextReplace, &ThisReplace);
-
-		++level;
-		goto loop;
-
-	      normal_match:
-		back = NextResult;
-	    }
-
-	    /* patch the &'s if needed */
-	    if (ThisReplace.type == C_REPLV) {
-		STRING *sval = new_STRING1(ThisMiddle, ThisMiddleLen);
-
-		replv_to_repl(&ThisReplace, sval);
-		free_STRING(sval);
+		j = (int) (where - input->str);
+		if (j < (int) input->len) {
+		    TRACE(("..emptied:"));
+		    TRACE_STRING2(input->str + j, 1);
+		    TRACE(("\n"));
+		    output->str[used++] = input->str[j];
+		}
 	    }
+	    skip0 = (howmuch != 0) ? (j + 1) : -1;
+	} else {
+	    have = (input->len - (size_t) j);
+	    TRACE(("..after match:%d:", (int) have));
+	    TRACE_STRING2(input->str + j, have);
+	    TRACE(("\n"));
+	    memcpy(output->str + used, input->str + j, have);
+	    used += have;
+	    break;
 	}
-
-	/* put the three pieces together */
-	ThisResult = new_STRING0(ThisFrontLen + string(&ThisReplace)->len + back->len);
-	TRACE(("new %s front '%.*s', middle '%.*s', back '%.*s'\n",
-	       indent(level),
-	       (int) ThisFrontLen, ThisFront,
-	       (int) string(&ThisReplace)->len, string(&ThisReplace)->str,
-	       (int) back->len, back->str));
-	in_sval = ThisResult->str;
-
-	if (ThisFrontLen) {
-	    memcpy(in_sval, ThisFront, ThisFrontLen);
-	    in_sval += ThisFrontLen;
-	}
-	if (string(&ThisReplace)->len) {
-	    memcpy(in_sval, string(&ThisReplace)->str, string(&ThisReplace)->len);
-	    in_sval += string(&ThisReplace)->len;
-	}
-	if (back->len)
-	    memcpy(in_sval, back->str, back->len);
-
-	/* cleanup, repl is freed by the caller */
-	free_STRING(back);
-	repl_destroy(&ThisReplace);
-
-    } else {
-	/* no match */
-	ThisResult = new_STRING1(ThisTarget, ThisTargetLen);
-	repl_destroy(&ThisReplace);
     }
 
-    switch (ThisBranch) {
-    case btEmpty:
-	--level;
-	goto empty_match;
-    case btNormal:
-	--level;
-	goto normal_match;
-    case btFinish:
-	break;
-    }
+    TRACE(("..input %d ->output %d\n",
+	   (int) input->len,
+	   (int) output->len));
 
-    return ThisResult;
+    repl_destroy(&xrepl);
+    if (output->len > used) {
+	buffer = output;
+	output = new_STRING1(output->str, used);
+	free_STRING(buffer);
+    }
+    TRACE(("..done gsub3\n"));
+    return output;
 }
 
 /* set up for call to gsub() */
 CELL *
-bi_gsub(CELL * sp)
+bi_gsub(CELL *sp)
 {
     CELL *cp;			/* pts at the replacement target */
     CELL sc;			/* copy of replacement target */
     CELL tc;			/* build the result here */
     STRING *result;
-#ifdef DEBUG_GSUB
-    STRING *resul2;
-#endif
-    size_t stack_needs;
-    int level = 0;
 
     TRACE_FUNC("bi_gsub", sp);
 
     sp -= 2;
+
     if (sp->type != C_RE)
 	cast_to_RE(sp);
     if ((sp + 1)->type != C_REPL && (sp + 1)->type != C_REPLV)
@@ -1472,209 +1358,35 @@ bi_gsub(CELL * sp)
     if (sc.type < C_STRING)
 	cast1_to_s(&sc);
 
-    stack_needs = (string(&sc)->len + 2) * 2;
+    TRACE(("..actual gsub args:\n"));
+    TRACE(("arg0: "));
+    TRACE_CELL(sp);
+    TRACE(("arg1: "));
+    TRACE_CELL(sp + 1);
+    TRACE(("arg2: "));
+    TRACE_CELL(&sc);
 
-    if (stack_needs > gsub_max) {
-	if (gsub_max) {
-	    zfree(gsub_stk, gsub_max * sizeof(GSUB_STK));
-	}
-	gsub_stk = zmalloc(stack_needs * sizeof(GSUB_STK));
-	gsub_max = stack_needs;
-    }
-#ifdef DEBUG_GSUB
-    {
-	STRING *target = new_STRING1(string(&sc)->str, string(&sc)->len);
-
-	ThisBranch = btFinish;
-	ThisEmptyOk = 1;
-	cellcpy(&ThisReplace, sp + 1);
-	ThisResult = 0;
-	ThisTarget = target->str;
-	ThisTargetLen = target->len;
-
-	resul2 = old_gsub(sp->ptr, 0);
-
-	TRACE(("OLD ->'%.*s'\n", resul2->len, resul2->str));
-	free_STRING(target);
-    }
-#endif
-
-    ThisBranch = btFinish;
-    ThisEmptyOk = 1;
-    cellcpy(&ThisReplace, sp + 1);
-    ThisResult = 0;
-    ThisTarget = string(&sc)->str;
-    ThisTargetLen = string(&sc)->len;
-
-    repl_cnt = 0;
-
-    result = new_gsub(sp->ptr, 0);
+    result = gsub3(sp->ptr, sp + 1, &sc);
     tc.ptr = (PTR) result;
 
-#ifdef DEBUG_GSUB
-    TRACE(("NEW ->'%.*s'\n", result->len, result->str));
-    if (result->len != resul2->len || memcmp(result->str, resul2->str, result->len))
-	TRACE(("OOPS\n"));
-#endif
-
     if (repl_cnt) {
 	tc.type = C_STRING;
 	slow_cell_assign(cp, &tc);
     }
-#ifdef NO_LEAKS
-    if (gsub_stk != 0) {
-	zfree(gsub_stk, stack_needs * sizeof(GSUB_STK));
-	gsub_stk = 0;
-	gsub_max = 0;
-    }
-#endif
+
+    sp->type = C_DOUBLE;
+    sp->dval = (double) repl_cnt;
+
+    TRACE(("Result: "));
+    TRACE_CELL(sp);
+    TRACE(("String: "));
+    TRACE_STRING(result);
+    TRACE(("\n"));
 
     /* cleanup */
     free_STRING(string(&sc));
     free_STRING(string(&tc));
     repl_destroy(sp + 1);
 
-    sp->type = C_DOUBLE;
-    sp->dval = (double) repl_cnt;
-
     return_CELL("bi_gsub", sp);
 }
-
-#else /* GSUB uses stack... */
-static unsigned repl_cnt;	/* number of global replacements */
-
-/* recursive global subsitution
-   dealing with empty matches makes this mildly painful
-
-   repl is always of type REPL or REPLV, destroyed by caller
-   flag is set if, match of empty string at front is OK
-*/
-
-static STRING *
-gsub(PTR re, CELL * repl, char *target, size_t target_len, int flag)
-{
-    char *front = 0, *middle;
-    STRING *back;
-    size_t front_len, middle_len;
-    STRING *ret_val;
-    CELL xrepl;			/* a copy of repl so we can change repl */
-
-    if (!(middle = REmatch(target, target_len, cast_to_re(re), &middle_len)))
-	return new_STRING(target);	/* no match */
-
-    cellcpy(&xrepl, repl);
-
-    if (!flag && middle_len == 0 && middle == target) {
-	/* match at front that's not allowed */
-
-	if (*target == 0) {	/* target is empty string */
-	    repl_destroy(&xrepl);
-	    null_str.ref_cnt++;
-	    return &null_str;
-	} else if (1 && isAnchored(re)) {
-	    repl_destroy(&xrepl);
-	    return new_STRING1(target, target_len);
-	} else {
-	    char xbuff[2];
-
-	    front_len = 0;
-	    /* make new repl with target[0] */
-	    repl_destroy(repl);
-	    --target_len;
-	    xbuff[0] = *target++;
-	    xbuff[1] = 0;
-	    repl->type = C_REPL;
-	    repl->ptr = (PTR) new_STRING(xbuff);
-	    back = gsub(re, &xrepl, target, target_len, 1);
-	}
-    } else {			/* a match that counts */
-	repl_cnt++;
-
-	front = target;
-	front_len = (unsigned) (middle - target);
-
-	if (front_len == target_len) {	/* matched back of target */
-	    back = &null_str;
-	    null_str.ref_cnt++;
-	} else {
-	    back = gsub(re,
-			&xrepl,
-			middle + middle_len,
-			target_len - (front_len + middle_len),
-			0);
-	}
-
-	/* patch the &'s if needed */
-	if (repl->type == C_REPLV) {
-	    STRING *sval = new_STRING0(middle_len);
-
-	    memcpy(sval->str, middle, middle_len);
-	    replv_to_repl(repl, sval);
-	    free_STRING(sval);
-	}
-    }
-
-    /* put the three pieces together */
-    ret_val = new_STRING0(front_len + string(repl)->len + back->len);
-    {
-	char *p = ret_val->str;
-
-	if (front_len) {
-	    memcpy(p, front, front_len);
-	    p += front_len;
-	}
-
-	if (string(repl)->len) {
-	    memcpy(p, string(repl)->str, string(repl)->len);
-	    p += string(repl)->len;
-	}
-	if (back->len)
-	    memcpy(p, back->str, back->len);
-    }
-
-    /* cleanup, repl is freed by the caller */
-    repl_destroy(&xrepl);
-    free_STRING(back);
-
-    return ret_val;
-}
-
-/* set up for call to gsub() */
-CELL *
-bi_gsub(CELL * sp)
-{
-    CELL *cp;			/* pts at the replacement target */
-    CELL sc;			/* copy of replacement target */
-    CELL tc;			/* build the result here */
-
-    TRACE_FUNC("bi_gsub", sp);
-
-    sp -= 2;
-    if (sp->type != C_RE)
-	cast_to_RE(sp);
-    if ((sp + 1)->type != C_REPL && (sp + 1)->type != C_REPLV)
-	cast_to_REPL(sp + 1);
-
-    cellcpy(&sc, cp = (CELL *) (sp + 2)->ptr);
-    if (sc.type < C_STRING)
-	cast1_to_s(&sc);
-
-    repl_cnt = 0;
-    tc.ptr = (PTR) gsub(sp->ptr, sp + 1, string(&sc)->str, string(&sc)->len, 1);
-
-    if (repl_cnt) {
-	tc.type = C_STRING;
-	slow_cell_assign(cp, &tc);
-    }
-
-    /* cleanup */
-    free_STRING(string(&sc));
-    free_STRING(string(&tc));
-    repl_destroy(sp + 1);
-
-    sp->type = C_DOUBLE;
-    sp->dval = (double) repl_cnt;
-
-    return_CELL("bi_gsub", sp);
-}
-#endif /* EXP_UNROLLED_GSUB */
